// HaxePreD: Haxe Preprocessor In D
// Written in the D programming language

/** 
A preprocessor for the Haxe language (http://www.haxe.org).

Author:
$(WEB www.semitwist.com, Nick Sabalausky)

Prerequisites for Building:
- [Required]    DMD 1.043 (LDC untested)
- [Required]    Tango 0.99.8
- [Required]    SemiTwistDTools trunk r86
- [Required]    Goldie v0.02
- [Recommended] Rebuild 0.76
DMD 1.044+ with Tango Trunk might work, but is untested.
Rebuild 0.78 might work, but is quirky and not recommended.

$(WEB http://www.dsource.org/projects/tango/wiki/Download, Bundled Tango/DMD download)
$(WEB http://www.digitalmars.com/d,                        D Programming Language)
$(WEB http://www.dsource.org/projects/semitwist,           SemiTwist D Tools)
$(WEB http://www.dsource.org/projects/goldie,              Goldie)
$(WEB http://www.dsource.org/projects/dsss/wiki/Rebuild,   Rebuild)

*/

module haxepred.lib;

import tango.core.Array;
import tango.core.Exception;
import tango.io.Console;
import tango.io.Stdout;
import tango.io.device.File;
import tango.io.stream.Lines;
import tango.io.stream.Buffered;
import Text = tango.text.Unicode;

import semitwist.util.all;

import goldie.all;

import haxepred.util;

class PreprocessException : Exception
{
	uint line;
	uint col;
	long index;
	long[] lineIndicies;
	
	this(HaxePreD hpd, char[] msg)
	{
		this.line = hpd.line;
		this.index = hpd.index;
		this.lineIndicies = hpd.lineIndicies;
		this.col = cast(uint)(index - lineIndicies[HaxePreD.lineAtIndex(lineIndicies, index)]);
		super("{0}: character {1} : {2}".sformat(line+1, col+1, msg));
	}
}

class ErrorDirectiveException : PreprocessException
{
	this(HaxePreD hpd)
	{
		super(hpd, "Not implemented for current platform");
	}
}

class InvalidDirectiveException : PreprocessException
{
	this(HaxePreD hpd, char[] directive)
	{
		super(hpd, "Invalid preprocessor directive '"~directive~"'");
	}
}

class NoMatchingIfException : PreprocessException
{
	this(HaxePreD hpd, char[] directive, bool directlyInElse=false)
	{
		super(hpd,
			"'{}' without matching '#if'{}"
				.sformat(
					directive,
					(directlyInElse? ", directly inside an '#else'":"")
				)
		);
	}
}

class UnterminatedIfException : PreprocessException
{
	this(HaxePreD hpd)
	{
		super(hpd, "Unterminated '#if'");
	}
}

class UnexpectedEOFException : PreprocessException
{
	this(HaxePreD hpd, char[] state)
	{
		super(hpd, "Unexpected EOF inside '{}'".sformat(state));
	}
}

class ConditionalExpressionException : PreprocessException
{
	Exception next;
	char[] expr;
	
	this(HaxePreD hpd, char[] expr, Exception next)
	{
		this.next = next;
		this.expr = expr;
		super(hpd, "Invalid conditional expression '{}': {}".sformat(expr, next));
	}
}

class InternalPreprocessException : PreprocessException
{
	this(HaxePreD hpd, char[] msg)
	{
		super(hpd, "Internal Error: "~msg);
	}
}

//TODO: Change err msg from "characters {end index}-{end index}" to "characters {start column}-{end column}"
//TODO: Report the location where errors start, rather than where they are detected.

class HaxePreD
{
	/// input:   Must be an InputStream (can be null) or a char[] filename.
	///          If null or "", stdin will be used.
	/// output:  Must be an OutputStream (can be null) or a char[] filename.
	///          If null or "", stdout will be used.
	/// defines: A list of identifiers to be considered to be defined.
	public static void preprocessStream(TIn, TOut)(TIn input, TOut output, char[][] defines=[])
	{
		static assert(
			is(TIn:InputStream) || is(TIn:char[]),
			"input must be an InputStream (can be null) or a char[] filename"
		);
		static assert(
			is(TOut:OutputStream) || is(TOut:char[]),
			"output must be an OutputStream (can be null) or a char[] filename"
		);

		scope hpd = new HaxePreD();

		hpd.setInput(input);
		hpd.setOutput(output);
		hpd.addBuffers();
		hpd.inBuf = new ReadAdjacentBuffer(hpd.inStream, 16, 1);
		hpd.defines = defines;
		hpd.process(true);
		hpd.outStream.flush();
		
		if(hpd.autoCloseInStream)
			hpd.inStream.close();
		if(hpd.autoCloseOutStream)
			hpd.outStream.close();

		lastLineIndicies = hpd.lineIndicies;
	}

	/// input:   Must be an InputStream (can be null) or a char[] filename.
	///          If null or "", stdin will be used.
	/// defines: A list of identifiers to be considered to be defined.
	/// Returns the preprocessed result.
	public static char[] preprocessStream(TIn, TOut)(TIn input, char[][] defines=[])
	{
		static assert(
			is(TIn:InputStream) || is(TIn:char[]),
			"input must be an InputStream (can be null) or a char[] filename"
		);

		scope hpd = new HaxePreD();

		hpd.setInput(input);
		hpd.addBuffers();
		hpd.inBuf = new ReadAdjacentBuffer(hpd.inStream, 16, 1);

		hpd.defines = defines;
		auto result = hpd.process(false);

		if(hpd.autoCloseInStream)
			hpd.inStream.close();

		lastLineIndicies = hpd.lineIndicies;
		return result;
	}
	
	/// input:   Haxe source code.
	/// defines: A list of identifiers to be considered to be defined.
	/// Returns the preprocessed result.
	public static char[] preprocessString(char[] input, char[][] defines=[])
	{
		scope hpd = new HaxePreD();

		hpd.inBuf = new ReadAdjacentBuffer(input);
		hpd.defines = defines;
		return hpd.process(false);
	}
	
	/// The indicies on which each line starts.
	/// Ex: Line #4 (zero-indexed) starts at index 'getLastLineIndicies()[4]'
	public static long[] getLastLineIndicies()
	{
		return lastLineIndicies;
	}
	
	private static long[] lastLineIndicies;
	
	public static int lineAtIndex(long[] lineIndicies, long index)
	{
		foreach_reverse(uint line, long lineStart; lineIndicies)
		{
			if(lineStart < index)
				return line;
		}
		return 0;
	}
	
	public static this()
	{
		exprLang = Goldie.loadLanguage("expr");
		goldLexer = new GoldLexer();
		goldParser = new GoldParser();
	}
	
	private static GoldLanguage exprLang;
	private static GoldLexer    goldLexer;
	private static GoldParser   goldParser;
	
	private InputStream inStream;
	private OutputStream outStream;
	private ReadAdjacentBuffer inBuf;
	
	private bool autoCloseInStream = false;
	private bool autoCloseOutStream = false;
	
	private this()
	{
	}

	private void setInput(InputStream input)
	{
		this.inStream = input? input : Cin.stream;
	}
	private void setInput(char[] input)
	{
		if(input == "")
			setInput(cast(InputStream)null);
		else
		{
			setInput( (new File(input)).input );
			autoCloseInStream = true;
		}
	}
	
	private void setOutput(OutputStream output)
	{
		this.outStream = output? output : Stdout;
	}
	private void setOutput(char[] output)
	{
		if(output == "")
			setOutput(cast(OutputStream)null);
		else
		{
			setOutput( (new File(output, File.WriteCreate)).output );
			autoCloseOutStream = true;
		}
	}
	
	private void addBuffers()
	{
		inStream = new BufferedInput(inStream);
		if(outStream)
			outStream = new BufferedOutput(outStream);
	}
	
	private TokGold parseCond(char[] condStr)
	{
		goldParser.process(goldLexer.process(condStr, exprLang), exprLang);
		return goldParser.parseTree;
	}

	private bool evalCond(char[] condStr)
	{
		TokGold pt;
		try
			pt = parseCond(condStr);
		catch(Exception e)
			throw new ConditionalExpressionException(this, condStr, e);

		return evalCond(pt);
	}
	
	private bool evalCond(TokGold tok)
	{
		//mixin(traceVal!("tok.name", "tok"));
		switch(tok.name)
		{
		case "Ident":
			return cast(bool)(defines.contains(tok.toString()));
		
		case "ValExpr":
			// Ident
			if(tok.length == 1)
				return evalCond(tok[0]);
				
			// '!' Ident
			if(tok.length == 2)
				return !evalCond(tok[1]);
			
			// '(' <OrExpr> ')'
			if(tok.length == 3)
				return evalCond(tok[1]);

			// '!' '(' <OrExpr> ')'
			return !evalCond(tok[2]);
		
		case "OrExpr":
			// <AndExpr>
			if(tok.length == 1)
				return evalCond(tok[0]);
			
			// <OrExpr> '||' <AndExpr>
			return evalCond(tok[0]) || evalCond(tok[2]);
			
		case "AndExpr":
			// <ValExpr>
			if(tok.length == 1)
				return evalCond(tok[0]);
			
			// <AndExpr> '&&' <ValExpr>
			return evalCond(tok[0]) && evalCond(tok[2]);
			
		default:
			throw new InternalPreprocessException(this,
				"Unexpected token type '{}' in conditional expression".sformat(tok.name)
			);
		}
	}
	
	private enum State
	{
		Normal,
		Keyword,
		LineComment,
		BlockComment,
		SingleQuote,
		DoubleQuote,
		Regex,
		CondExpr,
		CondExpr_Ident,
		CondExpr_Paren,
	}
	
	private static char[] stateToReadableString(State state)
	{
		char[][State] lookup =
		[
			State.Normal:         "Haxe Source"[],
			State.Keyword:        "Preprocessor Directive"[],
			State.LineComment:    "Line Comment"[],
			State.BlockComment:   "Block Comment"[],
			State.SingleQuote:    "Single Quote"[],
			State.DoubleQuote:    "Double Quote"[],
			State.Regex:          "Regular Expression"[],
			State.CondExpr:       "Conditional Expression"[],
			State.CondExpr_Ident: "Conditional Expression"[],
			State.CondExpr_Paren: "Conditional Expression"[],
		];
		
		return (state in lookup)? lookup[state] : "{{unknown: #{}}".sformat(state);
	}

	private char[][] defines;
	private State state;
	private long index;
	private uint line;
	private long[] lineIndicies;
	private char[] tokenStr;
	private bool eof;
	private char ch;
	private int parenDepth;

	private bool streamOut;
	private char[] outStr;
	
	private int charsWritten;
	
	private static const char[][] allDirectives = ["#if"[], "#else", "#elseif", "#end", "#error"];

	private struct IfState
	{
		bool isTrue;
		bool isInElse;
	}
	private IfState[] ifStateStack;
	
	private bool isTrue()
	{
		foreach(IfState state; ifStateStack)
			if(!state.isTrue)
				return false;
		
		return true;
	}
	
	private void reset()
	{
		state = State.Normal;
		index = 0;
		line = 0;
		lineIndicies = [0];
		eof = false;
		outStr = "";
		ifStateStack = [];
		charsWritten = 0;
		tokenStr = "";
		parenDepth = 0;
	}
	
	private void write(char outChar)
	{
		debug charsWritten++;
		
		if(streamOut)
		{
			auto ret = outStream.write(outChar~"");
			if(ret == -1)
				throw new IOException("Unexpected EOF in output stream");
			if(ret == 0)
				throw new IOException("Character couldn't be sent to output");
		}
		else
		{
			outStr ~= outChar;
		}
	}

	private void onDirective(char[] tok)
	{
		switch(tok)
		{
		case "#if":
			state = State.CondExpr;
			tokenStr = "";
			break;
			
		case "#else":
			if(ifStateStack.length == 0)
				throw new NoMatchingIfException(this, tok);
			if(ifStateStack[$-1].isInElse)
				throw new NoMatchingIfException(this, tok, true);
			ifStateStack[$-1].isInElse = true;
			ifStateStack[$-1].isTrue = !ifStateStack[$-1].isTrue;
			state = State.Normal;
			break;
			
		case "#elseif":
			if(ifStateStack.length == 0)
				throw new NoMatchingIfException(this, tok);
			if(ifStateStack[$-1].isInElse)
				throw new NoMatchingIfException(this, tok, true);
			ifStateStack[$-1].isTrue = true;
			state = State.CondExpr;
			tokenStr = "";
			break;
			
		case "#end":
			if(ifStateStack.length == 0)
				throw new NoMatchingIfException(this, tok);
			ifStateStack = ifStateStack[0..$-1]; // Pop
			state = State.Normal;
			break;
			
		case "#error":
			if(ifStateStack.length == 0 || ifStateStack[$-1].isTrue)
				throw new ErrorDirectiveException(this);
			break;
			
		default:
			throw new InternalPreprocessException(this, "Unexpected directive '{}'".sformat(tok));
		}
	}
	
	private char[] process(bool streamOut)
	{
		reset();
		this.streamOut = streamOut;
		while(!inBuf.eof)
		{
			ch = inBuf.curr;
			if(ch == '\n')
			{
				line++;
				lineIndicies ~= index;
			}
			
			switch(state)
			{
			case State.Normal:
				switch(ch)
				{
				case '#':
					state = State.Keyword;
					tokenStr = ch~"";
					write(' ');
					break;
					
				case '/':
					if(inBuf.prev.length > 0)
					{
						if(inBuf.prev[$-1] == '/')
							state = State.LineComment;
						else if(inBuf.prev[$-1] == '~')
							state = State.Regex;
					}

					write(ch);
					break;
					
				case '*':
					if(inBuf.prev.length > 0 && inBuf.prev[$-1] == '/')
							state = State.BlockComment;

					write(ch);
					break;
					
				case '\'':
					state = State.SingleQuote;
					write(ch);
					break;

				case '"':
					state = State.DoubleQuote;
					write(ch);
					break;
				
				default:
					if(isTrue || Text.isWhitespace(ch))
						write(ch);
					else
						write(' ');
					break;
				}
				break;
			
			case State.Keyword:
				write(' ');
				tokenStr ~= ch;

				if(inBuf.next.length == 0 || !Text.isLower(inBuf.next[0]))
				{
					if(allDirectives.contains(tokenStr))
						onDirective(tokenStr);
					else
						throw new InvalidDirectiveException(this, tokenStr);
				}
				break;

			case State.LineComment:
				write(ch);

				if(ch == '\n')
					state = State.Normal;
				break;

			case State.BlockComment:
				write(ch);

				if(ch == '/' && inBuf.prev.length > 0 && inBuf.prev[$-1] == '*')
					state = State.Normal;
				break;

			case State.SingleQuote:
				write(ch);

				if(ch == '\'' && (inBuf.prev.length == 0 || inBuf.prev[$-1] != '\\'))
					state = State.Normal;
				break;

			case State.DoubleQuote:
				write(ch);

				if(ch == '"' && (inBuf.prev.length == 0 || inBuf.prev[$-1] != '\\'))
					state = State.Normal;
				break;

			case State.Regex:
				write(ch);

				// Don't need to worry about the "igms" suffixes,
				// because State.Normal will handle them just fine anyway.
				if(ch == '/' && (inBuf.prev.length == 0 || inBuf.prev[$-1] != '\\'))
					state = State.Normal;
				break;

			case State.CondExpr:
				tokenStr ~= ch;
				
				if(Text.isWhitespace(ch))
					write(ch);
				else
				{
					write(' ');
					
					if(ch != '!')
					{
						if(ch == '(')
						{
							state = State.CondExpr_Paren;
							parenDepth = 1;
						}
						else
							state = State.CondExpr_Ident;
					}
				}
				break;
				
			case State.CondExpr_Ident:
				tokenStr ~= ch;

				bool identDone=false;
				if(Text.isWhitespace(ch))
				{
					write(ch);
					identDone = true;
				}
				else
				{
					write(' ');
					if(!Text.isLetterOrDigit(ch) && ch != '_')
						identDone = true;
				}
					
				if(identDone)
				{
					ifStateStack ~= IfState(evalCond(tokenStr), false); // Push
					state = State.Normal;
				}
				break;

			case State.CondExpr_Paren:
				tokenStr ~= ch;

				if(Text.isWhitespace(ch))
					write(ch);
				else
				{
					write(' ');
					if(ch == '(')
						parenDepth++;
					else if(ch == ')')
					{
						parenDepth--;
						if(parenDepth == 0)
						{
							ifStateStack ~= IfState(evalCond(tokenStr), false); // Push
							state = State.Normal;
						}
					}
				}
				break;

			default:
				throw new InternalPreprocessException(this, "Unknown state '{}'".sformat(state));
			}
			
			debug
			{
				if(charsWritten != 1)
					throw new InternalPreprocessException(this, "Wrong number of bytes generated (expected 1, got {})".sformat(charsWritten));
				charsWritten = 0;
			}
			
			index++;
			inBuf.advance();
		}
		
		if(state != State.Normal)
			throw new UnexpectedEOFException(this, stateToReadableString(state));
		
		if(ifStateStack.length != 0)
			throw new UnterminatedIfException(this);
			
		return outStr;
	}
}

