// HaxePreD: Haxe Preprocessor In D
// Written in the D programming language

/** 
A preprocessor for the Haxe language (http://www.haxe.org).

Author:
$(WEB www.semitwist.com, Nick Sabalausky)

Prerequisites for Building:
- [Required]    DMD 1.043 (LDC untested)
- [Required]    Tango 0.99.8
- [Required]    SemiTwistDTools trunk r86
- [Required]    Goldie v0.02
- [Recommended] Rebuild 0.76
DMD 1.044+ with Tango Trunk might work, but is untested.
Rebuild 0.78 might work, but is quirky and not recommended.

$(WEB http://www.dsource.org/projects/tango/wiki/Download, Bundled Tango/DMD download)
$(WEB http://www.digitalmars.com/d,                        D Programming Language)
$(WEB http://www.dsource.org/projects/semitwist,           SemiTwist D Tools)
$(WEB http://www.dsource.org/projects/goldie,              Goldie)
$(WEB http://www.dsource.org/projects/dsss/wiki/Rebuild,   Rebuild)

*/

module haxepred.main;

import tango.io.Stdout;
import tango.text.Util;

import semitwist.util.all;

import haxepred.cmd;
import haxepred.lib;

int main(char[][] args)
{
	flushAsserts();
	
	auto cmdsw = new CmdSwitches(args);
	if(cmdsw.shouldExit)
		return 1;
	
	try
		HaxePreD.preprocessStream(cmdsw.inFilename, cmdsw.outFilename, cmdsw.defines);
	catch(PreprocessException e)
	{
		Stdout.formatln("{}:{}", cmdsw.inFilename==""? "{stdin}" : cmdsw.inFilename, e.msg);
		return 1;
	}
	
	return 0;
}

unittest
{
	// BE VERY CAREFUL NOT TO CHANGE ANY WHITESPACE INSIDE THESE STRINGS!
	// (Line endings can be changed though, as long as the change is
	// consistent throughout.)
	const char[] src1 =
"#if AB
	AB
	#if XY
		AB && XY
	#else
		AB && !XY
	#end
	AB
#else
	!AB
	#if XY
		!AB && XY
	#else
		!AB && !XY
	#end
	!AB
#end";
	const char[] res1 =
"      
	  
	      
		        
	     
		         
	    
	  
     
	!AB
	      
		         
	     
		!AB && !XY
	    
	!AB
    ";
	const char[] res1_AB =
"      
	AB
	      
		        
	     
		AB && !XY
	    
	AB
     
	   
	      
		         
	     
		          
	    
	   
    ";
	const char[] res1_XY =
"      
	  
	      
		        
	     
		         
	    
	  
     
	!AB
	      
		!AB && XY
	     
		          
	    
	!AB
    ";
	const char[] res1_AB_XY =
"      
	AB
	      
		AB && XY
	     
		         
	    
	AB
     
	   
	      
		         
	     
		          
	    
	   
    ";

	mixin(deferAssert!(`HaxePreD.preprocessString(src1) == res1`));
	mixin(deferAssert!(`HaxePreD.preprocessString(src1, ["AB"[]]) == res1_AB`));
	mixin(deferAssert!(`HaxePreD.preprocessString(src1, ["XY"[]]) == res1_XY`));
	mixin(deferAssert!(`HaxePreD.preprocessString(src1, ["AB"[], "XY"]) == res1_AB_XY`));

	const char[] src2 =
"hi//#error
#if DEFTHIS
stuff
#else
foobar
#end
hey/*#error*/whoo
#if DEFTHIS
stuff
#else
foobar
#end";
	const char[] res2 =
"hi//#error
           
stuff
     
      
    
hey/*#error*/whoo
           
stuff
     
      
    ";
	mixin(deferAssert!(`HaxePreD.preprocessString(src2, ["DEFTHIS"[]]) == res2`));

	const char[] src3 =
"/*#error*/#if DEFTHIS
hi
#end";
	const char[] res3 =
"/*#error*/           
hi
    ";
//	mixin(deferAssert!(`HaxePreD.preprocessString(src3, ["DEFTHIS"[]]) == res3`));
//mixin(traceVal!(`HaxePreD.preprocessString(src3, ["DEFTHIS"[]])`));

	const char[] src4 =
`hi"#error"!
#if DEFTHIS
stuff
#else
foobar
#end
hey'#error'whoo
#if DEFTHIS
stuff
#else
foobar
#end`;
	const char[] res4 =
`hi"#error"!
           
stuff
     
      
    
hey'#error'whoo
           
stuff
     
      
    `;
	mixin(deferAssert!(`HaxePreD.preprocessString(src4, ["DEFTHIS"[]]) == res4`));
	
	const char[] src5 =
`"#error"#if DEFTHIS
hi
#end`;
	const char[] res5 =
`"#error"           
hi
    `;
	mixin(deferAssert!(`HaxePreD.preprocessString(src5, ["DEFTHIS"[]]) == res5`));

	const char[] src6 =
`'#error'#if DEFTHIS
hi
#end`;
	const char[] res6 =
`'#error'           
hi
    `;
	mixin(deferAssert!(`HaxePreD.preprocessString(src6, ["DEFTHIS"[]]) == res6`));

	const char[] src7 =
`~/#error/#if DEFTHIS
hi
#end`;
	const char[] res7 =
`~/#error/           
hi
    `;
	mixin(deferAssert!(`HaxePreD.preprocessString(src7, ["DEFTHIS"[]]) == res7`));

	const char[] src8 = `"#error\"#error"hi`;
	mixin(deferAssert!(`HaxePreD.preprocessString(src8) == src8`));

	const char[] src9 = `'#error\'#error'`;
	mixin(deferAssert!(`HaxePreD.preprocessString(src9) == src9`));

	const char[] src10 = `"hello"#error`;
	mixin(deferEnsureThrows!(`HaxePreD.preprocessString(src10);`, ErrorDirectiveException));

	mixin(deferEnsureThrows!(`HaxePreD.preprocessString("/*hello*/#error");`, ErrorDirectiveException));

	const char[] src11 = `#if AB hi #end`;
	mixin(deferAssert!(`HaxePreD.preprocessString(src11, ["AB"[]]).trim() == "hi"`));
	mixin(deferAssert!(`HaxePreD.preprocessString(src11, ["nAB"[]]).trim() == ""`));
	mixin(deferAssert!(`HaxePreD.preprocessString(src11, []).trim() == ""`));

	const char[] src12 = `#if (AB) hi #end`;
	mixin(deferAssert!(`HaxePreD.preprocessString(src12, ["AB"[]]).trim() == "hi"`));
	mixin(deferAssert!(`HaxePreD.preprocessString(src12, ["nAB"[]]).trim() == ""`));
	mixin(deferAssert!(`HaxePreD.preprocessString(src12, []).trim() == ""`));

	const char[] src13 = `#if AB #else hi #end`;
	mixin(deferAssert!(`HaxePreD.preprocessString(src13, ["AB"[]]).trim() == ""`));
	mixin(deferAssert!(`HaxePreD.preprocessString(src13, ["nAB"[]]).trim() == "hi"`));
	mixin(deferAssert!(`HaxePreD.preprocessString(src13, []).trim() == "hi"`));

	const char[] src14 = `#if !AB hi #end`;
	mixin(deferAssert!(`HaxePreD.preprocessString(src14, ["AB"[]]).trim() == ""`));
	mixin(deferAssert!(`HaxePreD.preprocessString(src14, ["nAB"[]]).trim() == "hi"`));
	mixin(deferAssert!(`HaxePreD.preprocessString(src14, []).trim() == "hi"`));

	const char[] src15 = `#if (!AB && ! CD || !WX && YZ ) hi #end`;
	mixin(deferAssert!(`HaxePreD.preprocessString(src15, ["nAB"[], "nCD", "nWX", "nYZ"]).trim() == "hi"`));
	mixin(deferAssert!(`HaxePreD.preprocessString(src15, ["nAB"[], "nCD", "nWX",  "YZ"]).trim() == "hi"`));
	mixin(deferAssert!(`HaxePreD.preprocessString(src15, ["nAB"[], "nCD",  "WX", "nYZ"]).trim() == "hi"`));
	mixin(deferAssert!(`HaxePreD.preprocessString(src15, ["nAB"[], "nCD",  "WX",  "YZ"]).trim() == "hi"`));

	mixin(deferAssert!(`HaxePreD.preprocessString(src15, ["nAB"[],  "CD", "nWX", "nYZ"]).trim() ==   ""`));
	mixin(deferAssert!(`HaxePreD.preprocessString(src15, ["nAB"[],  "CD", "nWX",  "YZ"]).trim() == "hi"`));
	mixin(deferAssert!(`HaxePreD.preprocessString(src15, ["nAB"[],  "CD",  "WX", "nYZ"]).trim() ==   ""`));
	mixin(deferAssert!(`HaxePreD.preprocessString(src15, ["nAB"[],  "CD",  "WX",  "YZ"]).trim() ==   ""`));

	mixin(deferAssert!(`HaxePreD.preprocessString(src15, [ "AB"[], "nCD", "nWX", "nYZ"]).trim() ==   ""`));
	mixin(deferAssert!(`HaxePreD.preprocessString(src15, [ "AB"[], "nCD", "nWX",  "YZ"]).trim() == "hi"`));
	mixin(deferAssert!(`HaxePreD.preprocessString(src15, [ "AB"[], "nCD",  "WX", "nYZ"]).trim() ==   ""`));
	mixin(deferAssert!(`HaxePreD.preprocessString(src15, [ "AB"[], "nCD",  "WX",  "YZ"]).trim() ==   ""`));

	mixin(deferAssert!(`HaxePreD.preprocessString(src15, [ "AB"[],  "CD", "nWX", "nYZ"]).trim() ==   ""`));
	mixin(deferAssert!(`HaxePreD.preprocessString(src15, [ "AB"[],  "CD", "nWX",  "YZ"]).trim() == "hi"`));
	mixin(deferAssert!(`HaxePreD.preprocessString(src15, [ "AB"[],  "CD",  "WX", "nYZ"]).trim() ==   ""`));
	mixin(deferAssert!(`HaxePreD.preprocessString(src15, [ "AB"[],  "CD",  "WX",  "YZ"]).trim() ==   ""`));

	const char[] src16 = `#if (!(AB || CD && !(!WX))) hi #end`;
	mixin(deferAssert!(`HaxePreD.preprocessString(src16, ["nAB"[], "nCD", "nWX"]).trim() == "hi"`));
	mixin(deferAssert!(`HaxePreD.preprocessString(src16, ["nAB"[], "nCD",  "WX"]).trim() == "hi"`));
	mixin(deferAssert!(`HaxePreD.preprocessString(src16, ["nAB"[],  "CD", "nWX"]).trim() == "hi"`));
	mixin(deferAssert!(`HaxePreD.preprocessString(src16, ["nAB"[],  "CD",  "WX"]).trim() ==   ""`));

	mixin(deferAssert!(`HaxePreD.preprocessString(src16, [ "AB"[], "nCD", "nWX"]).trim() ==   ""`));
	mixin(deferAssert!(`HaxePreD.preprocessString(src16, [ "AB"[], "nCD",  "WX"]).trim() ==   ""`));
	mixin(deferAssert!(`HaxePreD.preprocessString(src16, [ "AB"[],  "CD", "nWX"]).trim() ==   ""`));
	mixin(deferAssert!(`HaxePreD.preprocessString(src16, [ "AB"[],  "CD",  "WX"]).trim() ==   ""`));

	mixin(deferEnsureThrows!(`HaxePreD.preprocessString("#els");`, InvalidDirectiveException));
	mixin(deferEnsureThrows!(`HaxePreD.preprocessString("/*hello");`, UnexpectedEOFException));
	mixin(deferEnsureThrows!(`HaxePreD.preprocessString("//hello");`, UnexpectedEOFException));
	mixin(deferEnsureThrows!(`HaxePreD.preprocessString("#if (hi");`, UnexpectedEOFException));
	mixin(deferEnsureThrows!(`HaxePreD.preprocessString("\"hi");`, UnexpectedEOFException));
	mixin(deferEnsureThrows!(`HaxePreD.preprocessString("'hi");`, UnexpectedEOFException));
	mixin(deferEnsureThrows!(`HaxePreD.preprocessString("~/hi");`, UnexpectedEOFException));
	
	const char[] src17 =
`1
2
3
4
5
6234567890#error
7
8`;
	mixin(deferEnsureThrows!(`HaxePreD.preprocessString(src17);`, ErrorDirectiveException));
	try
		HaxePreD.preprocessString(src17);
	catch(ErrorDirectiveException e)
	{
		mixin(deferEnsure!(`e.line`, `_==5`));
		mixin(deferEnsure!(`e.col`, `_==16`));
	}
	catch(Object e)
		Stdout.formatln("Caught wrong exception");

}
