// HaxePreD: Haxe Preprocessor In D
// Written in the D programming language

module haxepred.util;

import tango.core.Exception;
import tango.io.device.Array;
import tango.io.device.Conduit;
import Math = tango.math.Math;
import tango.text.convert.Layout;

// Could I make this faster for streams by treating nextBuf and prevBuf as circular buffers?
// Probably just when next/prev aren't used.
class ReadAdjacentBuffer
{
	private InputStream inStream;
	private char[] inBuf;
	private char[] nextBuf;
	private char[] prevBuf;
	private char currCh;
	private bool streamEof;
	private int maxPrev;
	
	private char[] inString;
	private int index;
	
	this(InputStream inStream, int maxNext=16, int maxPrev=16)
	{
		reset(inStream, maxNext, maxPrev);
	}

	this(char[] inString)
	{
		reset(inString);
	}
	
	void reset(InputStream inStream, int maxNext=16, int maxPrev=16)
	{
		if(!inStream)
			throw new Exception("inStream is null");
		if(maxNext <= 0)
			throw new Exception("maxNext is not positive");
		if(maxPrev <= 0)
			throw new Exception("maxPrev is not positive");
			
		reset();

		this.inStream = inStream;
		this.maxPrev = maxPrev;
		streamEof=false;
		inBuf.length = 1;

		nextBuf.length = 0;
		nextBuf.length = maxNext;
		
		auto result = inStream.read(nextBuf);
		if(result == IOStream.Eof)
		{
			streamEof = true;
			nextBuf = "";
		}
		else
		{
			if(result < nextBuf.length)
				streamEof = true;
			
			nextBuf = nextBuf[0..result];
			advance();
			prevBuf = "";
		}
	}

	void reset(char[] inString)
	{
		reset();
		this.inString = inString;
	}
	
	private void reset()
	{
		inStream = null;
		inBuf = "";
		nextBuf = "";
		prevBuf = "";
		maxPrev = 0;
		streamEof = false;
		
		inString = "";
		index = 0;
		
		_eof = false;
	}
	
	char curr()
	{
		if(inStream)
			return currCh;
		else
			return inString[index];
	}
	
	bool _eof;
	bool eof()
	{
		return _eof;
	}
	
	// False if EOF
	bool advance()
	{
		if(!eof)
		{
			if(inStream)
			{
				if(!streamEof)
				{
					auto result = inStream.read(inBuf);
					if(result == IOStream.Eof || result == 0)
						streamEof = true;
					else
						nextBuf = nextBuf~inBuf[0];
				}
				if(nextBuf.length > 0)
				{
					if(prevBuf.length == maxPrev)
						prevBuf = prevBuf[1..$];
					prevBuf ~= currCh;
					currCh = nextBuf[0];
					nextBuf = nextBuf[1..$];
				}
				else
					_eof = true;
			}
			else
			{
				index++;
				if(index == inString.length)
					_eof = true;
			}
		}

		return _eof;
	}
	
	char[] next()
	{
		if(inStream)
			return nextBuf;
		else
			return inString[index+1..$];
	}

	char[] prev()
	{
		if(inStream)
			return prevBuf;
		else
			return inString[0..index];
	}
}

unittest
{
	auto b = new ReadAdjacentBuffer("0123456789");
	
	assert(!b.eof);
	assert(b.curr == '0');
	assert(b.next == "123456789");
	assert(b.prev == "");

	b.advance();
	b.advance();

	assert(!b.eof);
	assert(b.curr == '2');
	assert(b.next == "3456789");
	assert(b.prev == "01");

	b.advance();
	b.advance();
	b.advance();
	b.advance();

	assert(!b.eof);
	assert(b.curr == '6');
	assert(b.next == "789");
	assert(b.prev == "012345");

	b.advance();
	b.advance();
	b.advance();
	assert(!b.eof);
	b.advance();
	assert(b.eof);
	
	b = new ReadAdjacentBuffer("012");
	assert(!b.eof);
	assert(b.curr == '0');
	assert(b.next == "12");
	assert(b.prev == "");
	b.advance();
	assert(!b.eof);
	assert(b.curr == '1');
	assert(b.prev == "0");
	
	b = new ReadAdjacentBuffer(new Array("0123456789"), 3, 3);
	assert(!b.eof);
	assert(b.curr == '0');
	assert(b.next == "123");
	assert(b.prev == "");
	
	b.advance();
	b.advance();
	assert(!b.eof);
	assert(b.curr == '2');
	assert(b.next == "345");
	assert(b.prev == "01");
	
	b.advance();
	b.advance();
	b.advance();
	assert(!b.eof);
	assert(b.curr == '5');
	assert(b.next == "678");
	assert(b.prev == "234");

	b.advance();
	b.advance();
	assert(!b.eof);
	assert(b.curr == '7');
	assert(b.next == "89");
	assert(b.prev == "456");

	b.advance();
	b.advance();
	assert(!b.eof);
	b.advance();
	assert(b.eof);
	
	b = new ReadAdjacentBuffer(new Array("012"), 5, 5);
	assert(!b.eof);
	assert(b.curr == '0');
	assert(b.next == "12");
	assert(b.prev == "");
	
	b.advance();
	b.advance();
	assert(!b.eof);
	b.advance();
	assert(b.eof);
	assert(b.curr == '2');
	assert(b.next == "");
	assert(b.prev == "01");
}
